@@ -229,7 +229,10 @@ singlebyte:
       else
 	{
 	  m = mbrtowc (&wc, string + start, end - start, &state);
-	  if (MB_INVALIDCH (m) || m == 1)
+	  /* Have to go through wide case conversion even for single-byte
+	     chars, to accommodate single-byte characters where the
+	     corresponding upper or lower case equivalent is multibyte. */
+	  if (MB_INVALIDCH (m))
 	    {
 	      wc = (unsigned char)string[start];
 	      goto singlebyte;
